/*
 * Copyright (c) 2017 Adrian Siekierka
 *
 * This file is part of Unlimited Chisel Works.
 *
 * Unlimited Chisel Works is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Unlimited Chisel Works is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Unlimited Chisel Works.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.ucw;

import com.google.common.collect.ImmutableList;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.state.IBlockState;
import net.minecraft.client.renderer.block.model.BakedQuad;
import net.minecraft.client.renderer.block.model.IBakedModel;
import net.minecraft.client.renderer.block.model.ModelResourceLocation;
import net.minecraft.client.renderer.texture.TextureAtlasSprite;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ResourceLocation;
import net.minecraftforge.client.model.IModel;
import net.minecraftforge.common.model.TRSRTransformation;

import java.awt.image.BufferedImage;
import java.util.Collection;
import java.util.List;

public final class UCWMagic {
	private static class FakeSprite extends TextureAtlasSprite {
		protected FakeSprite(ResourceLocation spriteName) {
			super(spriteName.toString());
			func_110966_b(16);
			func_110969_c(16);
			func_110971_a(16, 16, 0, 0, false);
			func_110968_a(fakeTextureFrameList);
		}
	}

	private static final List<int[][]> fakeTextureFrameList;
	public static final BufferedImage missingNo;

	static {
		fakeTextureFrameList = ImmutableList.of(new int[1][256]);

		missingNo = new BufferedImage(16, 16, BufferedImage.TYPE_INT_ARGB);
		for (int i = 0; i < 256; i++) {
			missingNo.setRGB(i & 15, i >> 4, ((i ^ (i >> 4)) & 8) != 0 ? 0xFF000000 : 0xFFFF00FF);
		}
	}

	private UCWMagic() {

	}

	public static ResourceLocation getLocation(IBlockState state, ModelResourceLocation location, IModel model) {
		String domain = location.func_110624_b();

		if ("forestry".equals(domain)) {
			String[] name = state.func_177230_c().getRegistryName().func_110623_a().split("\\.", 2);
			IProperty variantProp = state.func_177230_c().func_176194_O().func_185920_a("variant");
			if (variantProp != null) {
				String variant = variantProp.func_177702_a(state.func_177229_b(variantProp));
				if (name.length == 2 && ("planks".equals(name[0]))) {
					return new ResourceLocation("forestry", "blocks/wood/" + name[0] + "." + variant);
				}
			}
		} else if ("extratrees".equals(domain)) {
			String[] name = state.func_177230_c().getRegistryName().func_110623_a().split("\\.", 2);
			IProperty variantProp = state.func_177230_c().func_176194_O().func_185920_a("variant");
			if (variantProp != null) {
				String variant = variantProp.func_177702_a(state.func_177229_b(variantProp));
				if (name.length == 2 && ("planks".equals(name[0]))) {
					return new ResourceLocation("extratrees", "blocks/planks/" + variant);
				}
			}
		}


		if (model.getTextures().size() == 1) {
			return model.getTextures().iterator().next();
		} else {
			try {
				// some mods have a skewed particle texture
				IBakedModel bakedModel = model.bake(TRSRTransformation.identity(), DefaultVertexFormats.field_176599_b, FakeSprite::new);
				Collection<BakedQuad> quadList = bakedModel.func_188616_a(state, EnumFacing.NORTH, 0);
				if (quadList.size() > 0) {
					return new ResourceLocation(quadList.iterator().next().func_187508_a().func_94215_i());
				}
			} catch (Exception e) {

			}

			// fallback
			return model.getTextures().iterator().next();
		}
	}

	private static float[] toLAB(int rgb) {
		return UCWColorspaceUtils.XYZtoLAB(UCWColorspaceUtils.sRGBtoXYZ(UCWColorspaceUtils.fromInt(rgb)));
	}

	private static int fromLAB(float[] lab) {
		return UCWColorspaceUtils.asInt(UCWColorspaceUtils.XYZtosRGB(UCWColorspaceUtils.LABtoXYZ(lab)));
	}

	private static float[] calculateContrast(int[] data) {
		float[] contrast = new float[] { Float.MAX_VALUE, Float.MIN_VALUE };

		for (int i : data) {
			float[] d = toLAB(i);
			if (contrast[0] > d[0]) contrast[0] = d[0];
			if (contrast[1] < d[0]) contrast[1] = d[0];
		}

		contrast[1] -= contrast[0];
		return contrast;
	}

	public static int[][] createBaseForColorMultiplier(TextureAtlasSprite texture, boolean keepTinting) {
		int[][] frames = new int[texture.func_110970_k()][texture.func_94211_a() * texture.func_94216_b()];

		int divider = texture.func_94211_a() * texture.func_94216_b() * texture.func_110970_k();
		double avgLuma = 0.0;

		for (int i = 0; i < texture.func_110970_k(); i++) {
			for (int v : texture.func_147965_a(i)[0]) {
				float[] hd = toLAB(v);
				avgLuma += hd[0];
			}
		}
		avgLuma /= divider;

		for (int i = 0; i < texture.func_110970_k(); i++) {
			int[] data = texture.func_147965_a(i)[0];
			for (int j = 0; j < texture.func_94211_a() * texture.func_94216_b(); j++) {
				float[] hd = toLAB(data[j]);

				frames[i][j] = (data[j] & 0xFF000000) | fromLAB(new float[]{(float) (hd[0] / avgLuma * 100.0), keepTinting ? hd[1] : 0, keepTinting ? hd[2] : 0});
			}
		}

		return frames;
	}

	public static int[] transform(TextureAtlasSprite sprite, int frame, TextureAtlasSprite from, TextureAtlasSprite overlay, TextureAtlasSprite basedUpon, UCWBlockRule.BlendMode mode) {
		int[] texture = sprite.func_147965_a(frame)[0];
		int width = sprite.func_94211_a();
		int height = sprite.func_94216_b();
		float[] contrastFrom = calculateContrast(from.func_147965_a(0)[0]);
		float[] contrastBasedUpon = calculateContrast(basedUpon.func_147965_a(0)[0]);
		double avgA = 0.0, avgB = 0.0;
		double[] rangeA = new double[2], rangeB = new double[2];
		int[] rangeDiv = new int[2];

		if (mode == UCWBlockRule.BlendMode.PLANK) {
			for (int i : from.func_147965_a(0)[0]) {
				float[] hd = toLAB(i);
				double normV = (double) (hd[0] - contrastFrom[0]) / contrastFrom[1];
				if (normV < 0.5) {
					rangeA[0] += hd[1];
					rangeB[0] += hd[2];
					rangeDiv[0]++;
				} else {
					rangeA[1] += hd[1];
					rangeB[1] += hd[2];
					rangeDiv[1]++;
				}
			}

			if (rangeDiv[0] > 0) {
				rangeA[0] /= rangeDiv[0];
				rangeB[0] /= rangeDiv[0];
			}

			if (rangeDiv[1] > 0) {
				rangeA[1] /= rangeDiv[1];
				rangeB[1] /= rangeDiv[1];
			}
		}

		if (mode == UCWBlockRule.BlendMode.BLEND) {
			for (int i : from.func_147965_a(0)[0]) {
				float[] data = toLAB(i);
				avgA += data[1];
				avgB += data[2];
			}

			avgA /= from.func_94211_a() * from.func_94216_b();
			avgB /= from.func_94211_a() * from.func_94216_b();
		}

		int[] texData = new int[texture.length];
		for (int iy = 0; iy < height; iy++) {
			for (int ix = 0; ix < width; ix++) {
				int i = iy*width+ix;
				int it = texture[i];
				int ibu = overlay.func_147965_a(0)[0][(iy % from.func_94216_b())*from.func_94211_a() + (ix % from.func_94211_a())];

				float[] hsbTex = toLAB(it);
				float[] hsbBu = toLAB(ibu);
				double normV;
				float v;
				if (contrastBasedUpon[1] != 0.0) {
					normV = (double) (hsbTex[0] - contrastBasedUpon[0]) / contrastBasedUpon[1];
				} else {
					normV = 0.5f;
				}
				v = (float) ((normV * contrastFrom[1]) + contrastFrom[0]);

				if (mode == UCWBlockRule.BlendMode.BLEND) {
					hsbBu[1] = (float) avgA;
					hsbBu[2] = (float) avgB;
				} else if (mode == UCWBlockRule.BlendMode.PLANK) {
					double nv2 = normV;
					if (nv2 < 0) nv2 = 0;
					else if (nv2 > 1) nv2 = 1;
					hsbBu[1] = (float) ((rangeA[1] * nv2) + (rangeA[0] * (1 - nv2)));
					hsbBu[2] = (float) ((rangeB[1] * nv2) + (rangeB[0] * (1 - nv2)));
				}

				if (v < 0) v = 0;
				else if (v > 100) v = 100;
				texData[i] = (it & 0xFF000000) | fromLAB(new float[]{v, hsbBu[1], hsbBu[2]});
			}
		}
		return texData;
	}
}
